# Copyright (c) Jon Parise <jon@indelible.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Literal
from typing import Mapping
from typing import Optional
from typing import TypedDict

from .chars import Rows

__all__ = ["Component", "Props", "Style"]

#: Map of dynamic properties that can be injected into message templates.
Props = Mapping[str, str]

#: Horizontal alignment options.
Justification = Literal["left", "right", "center", "justified"]

#: Vertical alignment options.
Alignment = Literal["top", "bottom", "center"]


class Position(TypedDict):
    """A Position defines an absolute position on a board."""

    x: int  #: X coordinate
    y: int  #: Y coordinate


class Style(TypedDict, total=False):
    """A Style defines a set of values that will be applied to a
    :py:class:`Component`.
    """

    #: The height of the current component (1-6). Defaults to 6.
    height: int

    #: The width of the current component (1-22). Defaults to 22.
    width: int

    #: The horizontal alignment of the message. Defaults to ``left``.
    justify: Justification

    #: The vertical alignment of a message. Defaults to ``top``.
    align: Alignment

    #: An exact position on the board. Bounded by ``width`` and ``height``.
    absolutePosition: Position


# Component is defined as a simple class so that we can provide good typing and
# a nice constructor.
#
# A TypedDict would correctly represent the API shape of a Component, but it
# doesn't support making the 'template' field "required" and the 'style' field
# "optional" until Python 3.11+, nor can we provide a custom constructor or
# "builder" class method.
#
# Using a dataclass results in an overall more complex class that also can't
# represent an "optional" field as "not present". (Dataclass optional fields
# are always set to their default values when an explicit value isn't given.)
class Component:
    """A Component defines a template with optional :py:class:`Style` values.

    Template is a message template string. It can contain any supported
    Vestaboard character as well as valid character code sequences inside curly
    braces, such as ``{5}`` or ``{65}`` and interpolated :py:class:`Props`
    using double bracket notation (``{{propName}}``). Any lowercase letters
    will be cast to uppercase. A new row can be forced by inserting a newline
    (``\\n``) sequence.

    Alternatively, a ``raw_characters`` character array can be provided to set
    the initial (background) state of the component. ``raw_characters`` takes
    precedence over ``template``.

    A ``style`` dictionary can be provided. Individual :py:class:`Style` values
    can also be given as keyword arguments (``height``, ``width``, ``justify``,
    ``align``, ``absolute_position``), and they will override any values from
    the ``style`` dictionary.
    """

    __slots__ = ["raw_characters", "style", "template"]

    def __init__(
        self,
        template: Optional[str] = None,
        raw_characters: Optional[Rows] = None,
        style: Optional[Style] = None,
        *,
        height: Optional[int] = None,
        width: Optional[int] = None,
        justify: Optional[Justification] = None,
        align: Optional[Alignment] = None,
        absolute_position: Optional[Position] = None,
    ):
        if not (template is not None or raw_characters is not None):
            raise ValueError("expected template or raw_characters")

        self.template = template
        self.raw_characters = raw_characters
        self.style: Style = style or {}
        if height is not None:
            self.style["height"] = height
        if width is not None:
            self.style["width"] = width
        if justify is not None:
            self.style["justify"] = justify
        if align is not None:
            self.style["align"] = align
        if absolute_position is not None:
            self.style["absolutePosition"] = absolute_position

    def asdict(self) -> Dict[str, Any]:
        """Returns the component's JSON dictionary representation."""
        d: Dict[str, Any] = {}
        if self.raw_characters is not None:
            d["rawCharacters"] = self.raw_characters
        elif self.template is not None:
            d["template"] = self.template
        if self.style:
            d["style"] = self.style
        return d



from .chars import Color
from .chars import encode
from .chars import encode_row
from .chars import encode_text
from .chars import pprint
from .clients import Client
from .clients import LocalClient
from .clients import ReadWriteClient
from .clients import SubscriptionClient
from .clients import VBMLClient
from .http import HTTPError

__all__ = (
    "Client",
    "Color",
    "HTTPError",
    "LocalClient",
    "ReadWriteClient",
    "SubscriptionClient",
    "VBMLClient",
    "encode",
    "encode_row",
    "encode_text",
    "pprint",
)

__version__ = "0.13.0"



import pytest

from vesta.vbml import Component


class TestComponent:
    def test_template(self):
        assert Component("template").asdict() == {"template": "template"}

    def test_raw_characters(self):
        assert Component(raw_characters=[]).asdict() == {"rawCharacters": []}

    def test_template_and_raw_characters(self):
        assert Component("t", raw_characters=[]).asdict() == {"rawCharacters": []}

    def test_template_or_raw_characters_required(self):
        with pytest.raises(ValueError, match="expected template or raw_characters"):
            Component()

    def test_style(self):
        assert Component("template", style={}).asdict() == {"template": "template"}
        assert Component(
            "template",
            style={"height": 4, "width": 10, "justify": "left", "align": "top"},
        ).asdict() == {
            "template": "template",
            "style": {"height": 4, "width": 10, "justify": "left", "align": "top"},
        }

    def test_height(self):
        assert Component("template", height=4).asdict() == {
            "template": "template",
            "style": {"height": 4},
        }

    def test_width(self):
        assert Component("template", width=10).asdict() == {
            "template": "template",
            "style": {"width": 10},
        }

    def test_justify(self):
        assert Component("template", justify="left").asdict() == {
            "template": "template",
            "style": {"justify": "left"},
        }

    def test_align(self):
        assert Component("template", align="top").asdict() == {
            "template": "template",
            "style": {"align": "top"},
        }

    def test_absolute_position(self):
        assert Component("template", absolute_position={"x": 1, "y": 2}).asdict() == {
            "template": "template",
            "style": {"absolutePosition": {"x": 1, "y": 2}},
        }



# Copyright (c) Jon Parise <jon@indelible.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import urllib.error
import urllib.request
from json import JSONDecodeError
from json import dumps as json_dumps
from json import loads as json_loads
from typing import Any
from typing import Dict
from typing import Literal
from typing import Mapping
from typing import NamedTuple
from typing import Optional
from urllib.parse import urljoin

# Re-export as a convenience
HTTPError = urllib.error.HTTPError


class Response(NamedTuple):
    """HTTP response tuple."""

    status: int
    body: bytes
    headers: Dict[str, str]

    def json(self, default: Any = None):
        try:
            return json_loads(self.body)
        except JSONDecodeError:
            return default


class Client:
    """HTTP client with connection pooling and session management.

    Uses urllib.request for HTTP operations with automatic connection reuse.
    """

    def __init__(
        self,
        base_url: str = "",
        headers: Optional[Mapping[str, str]] = None,
    ):
        self.base_url = base_url
        self.headers: Dict[str, str] = dict(headers) if headers else {}
        self._opener = urllib.request.build_opener()

    def request(
        self,
        method: Literal["GET", "POST"],
        url: str,
        *,
        json: Optional[Any] = None,
        headers: Optional[Mapping[str, str]] = None,
    ) -> Response:
        """Execute HTTP request and return Response.

        Raises urllib.error.HTTPError for 4xx/5xx responses.
        Raises urllib.error.URLError for connection errors.
        """
        full_url = urljoin(self.base_url, url.lstrip("/"))

        request_headers = dict(self.headers)
        if headers:
            request_headers.update(headers)

        data = None
        if json is not None:
            data = json_dumps(json, separators=(",", ":")).encode("utf-8")
            request_headers["Content-Type"] = "application/json"

        request = urllib.request.Request(
            full_url,
            data=data,
            headers=request_headers,
            method=method,
        )

        with self._opener.open(request) as response:
            return Response(
                response.getcode(),
                response.read(),
                dict(response.headers.items()),
            )



from __future__ import annotations

from json import dumps as json_dumps

import pytest

from vesta.chars import COLS
from vesta.chars import ROWS
from vesta.clients import Client
from vesta.clients import LocalClient
from vesta.clients import ReadWriteClient
from vesta.clients import SubscriptionClient
from vesta.clients import VBMLClient
from vesta.http import Response
from vesta.vbml import Component


@pytest.fixture
def mock_response(monkeypatch):
    """Fixture that returns a factory for mocking HTTP responses."""

    def _mock(client, status_code=200, body=None, json=None):
        last_request = {}

        if body is not None:
            response_body = body
        elif json is not None:
            response_body = json_dumps(json).encode("utf-8")
        else:
            response_body = b"{}"

        def fake_request(method, url, **kwargs):
            last_request.update({"method": method, "url": url, **kwargs})
            return Response(status_code, response_body, {})

        monkeypatch.setattr(client.http, "request", fake_request)
        return last_request

    return _mock


@pytest.fixture
def client():
    return Client("key", "secret")


@pytest.fixture
def local_client():
    return LocalClient("key")


@pytest.fixture
def rw_client():
    return ReadWriteClient("key")


@pytest.fixture
def subscription_client():
    return SubscriptionClient("key", "secret")


@pytest.fixture
def vbml_client():
    return VBMLClient()


@pytest.mark.filterwarnings("ignore:Vestaboard has deprecated the Platform API")
class TestClient:
    def test_base_url(self):
        base_url = "https://www.example.com"
        client = Client("key", "secret", base_url=base_url)
        assert client.http.base_url == base_url
        assert base_url in repr(client)

    def test_headers(self):
        client = Client("key", "secret", headers={"User-Agent": "Vesta"})
        assert client.http.headers["X-Vestaboard-Api-Key"] == "key"
        assert client.http.headers["X-Vestaboard-Api-Secret"] == "secret"
        assert client.http.headers["User-Agent"] == "Vesta"

    def test_get_subscriptions(self, client: Client, mock_response):
        subscriptions = [True]
        mock_response(client, json={"subscriptions": subscriptions})
        assert client.get_subscriptions() == subscriptions

    def test_get_viewer(self, client: Client, mock_response):
        viewer = {"_id": 1}
        mock_response(client, json=viewer)
        assert client.get_viewer() == viewer

    def test_post_message_text(self, client: Client, mock_response):
        text = "abc"
        mock = mock_response(client)
        client.post_message("sub_id", text)

        assert mock["url"] == "/subscriptions/sub_id/message"
        assert mock["json"] == {"text": text}

    def test_post_message_list(self, client: Client, mock_response):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(client)
        client.post_message("sub_id", chars)

        assert mock["url"] == "/subscriptions/sub_id/message"
        assert mock["json"] == {"characters": chars}

    def test_post_message_list_dimensions(self, client: Client):
        with pytest.raises(ValueError, match=rf"expected a \({COLS}, {ROWS}\) array"):
            client.post_message("sub_id", [])

    def test_post_message_type(self, client: Client):
        with pytest.raises(TypeError, match=r"unsupported message type"):
            client.post_message("sub_id", True)  # type: ignore


class TestLocalClient:
    def test_base_url(self):
        base_url = "http://example.local"
        local_client = LocalClient("key", base_url=base_url)
        assert local_client.http.base_url == base_url
        assert base_url in repr(local_client)

    def test_api_key_init(self):
        local_api_key = "key"
        local_client = LocalClient(local_api_key)
        assert local_client.enabled
        assert local_client.api_key == local_api_key

    def test_api_key_property(self):
        local_client = LocalClient()
        assert not local_client.enabled
        assert local_client.api_key is None

        local_api_key = "key"
        local_client.api_key = local_api_key
        assert local_client.enabled
        assert local_client.api_key == local_api_key

    def test_enable(self, mock_response):
        local_client = LocalClient()
        local_api_key = "key"
        mock = mock_response(local_client, json={"apiKey": local_api_key})
        rv = local_client.enable("enablement_token")
        assert rv == local_api_key
        assert local_client.api_key == local_api_key
        assert local_client.enabled

        assert mock["url"] == "/local-api/enablement"
        assert (
            mock["headers"]["X-Vestaboard-Local-Api-Enablement-Token"]
            == "enablement_token"
        )

    def test_enable_failure(self, mock_response):
        local_client = LocalClient()
        mock_response(local_client)
        rv = local_client.enable("enablement_token")
        assert rv is None
        assert not local_client.api_key
        assert not local_client.enabled

    def test_not_enabled(self):
        local_client = LocalClient()
        with pytest.raises(RuntimeError, match=r"Local API has not been enabled"):
            local_client.read_message()
        with pytest.raises(RuntimeError, match=r"Local API has not been enabled"):
            local_client.write_message([])

    def test_read_message(self, local_client: LocalClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock_response(local_client, json={"message": chars})
        message = local_client.read_message()
        assert message == chars

    def test_read_message_empty(self, local_client: LocalClient, mock_response):
        mock_response(local_client)
        message = local_client.read_message()
        assert message is None

    def test_write_message(self, local_client: LocalClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(local_client, status_code=201, body=b"")
        assert local_client.write_message(chars)

        assert mock["url"] == "/local-api/message"
        assert mock["json"] == chars

    def test_write_message_dimensions(self, local_client: LocalClient):
        with pytest.raises(ValueError, match=rf"expected a \({COLS}, {ROWS}\) array"):
            local_client.write_message([])


class TestReadWriteClient:
    def test_base_url(self):
        base_url = "http://example.local"
        rw_client = ReadWriteClient("key", base_url=base_url)
        assert rw_client.http.base_url == base_url
        assert base_url in repr(rw_client)

    def test_headers(self):
        client = ReadWriteClient("key", headers={"User-Agent": "Vesta"})
        assert client.http.headers["X-Vestaboard-Read-Write-Key"] == "key"
        assert client.http.headers["User-Agent"] == "Vesta"

    def test_read_message(self, rw_client: ReadWriteClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock_response(rw_client, json={"currentMessage": {"layout": json_dumps(chars)}})
        message = rw_client.read_message()
        assert message == chars

    def test_read_message_empty(self, rw_client: ReadWriteClient, mock_response):
        mock_response(rw_client)
        message = rw_client.read_message()
        assert message is None

    def test_read_message_empty_layout(self, rw_client: ReadWriteClient, mock_response):
        mock_response(rw_client, json={"currentMessage": {"layout": ""}})
        message = rw_client.read_message()
        assert message is None

    def test_write_message_text(self, rw_client: ReadWriteClient, mock_response):
        text = "abc"
        mock_response(rw_client)
        rw_client.write_message(text)

    def test_write_message_list(self, rw_client: ReadWriteClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(rw_client)
        assert rw_client.write_message(chars)

        assert mock["url"] == ""
        assert mock["json"] == chars

    def test_write_message_list_dimensions(self, rw_client: ReadWriteClient):
        with pytest.raises(ValueError, match=rf"expected a \({COLS}, {ROWS}\) array"):
            rw_client.write_message([])

    def test_write_message_type(self, rw_client: ReadWriteClient):
        with pytest.raises(TypeError, match=r"unsupported message type"):
            rw_client.write_message(True)  # type: ignore


class TestSubscriptionClient:
    def test_base_url(self):
        base_url = "https://www.example.com"
        client = SubscriptionClient("key", "secret", base_url=base_url)
        assert client.http.base_url == base_url
        assert base_url in repr(client)

    def test_headers(self):
        client = SubscriptionClient("key", "secret", headers={"User-Agent": "Vesta"})
        assert client.http.headers["x-vestaboard-api-key"] == "key"
        assert client.http.headers["x-vestaboard-api-secret"] == "secret"
        assert client.http.headers["User-Agent"] == "Vesta"

    def test_get_subscriptions(
        self, subscription_client: SubscriptionClient, mock_response
    ):
        subscriptions = [True]
        mock_response(subscription_client, json=subscriptions)
        assert subscription_client.get_subscriptions() == subscriptions

    def test_send_message_text(
        self, subscription_client: SubscriptionClient, mock_response
    ):
        text = "abc"
        mock = mock_response(subscription_client)
        subscription_client.send_message("sub_id", text)

        assert mock["url"] == "/subscriptions/sub_id/message"
        assert mock["json"] == {"text": text}

    def test_send_message_list(
        self, subscription_client: SubscriptionClient, mock_response
    ):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(subscription_client)
        subscription_client.send_message("sub_id", chars)

        assert mock["url"] == "/subscriptions/sub_id/message"
        assert mock["json"] == {"characters": chars}

    def test_send_message_list_dimensions(
        self, subscription_client: SubscriptionClient
    ):
        with pytest.raises(ValueError, match=rf"expected a \({COLS}, {ROWS}\) array"):
            subscription_client.send_message("sub_id", [])

    def test_send_message_type(self, subscription_client: SubscriptionClient):
        with pytest.raises(TypeError, match=r"unsupported message type"):
            subscription_client.send_message("sub_id", True)  # type: ignore


class TestVBMLClient:
    def test_base_url(self):
        base_url = "http://example.local"
        client = VBMLClient(base_url=base_url)
        assert client.http.base_url == base_url
        assert base_url in repr(client)

    def test_headers(self):
        client = VBMLClient(headers={"User-Agent": "Vesta"})
        assert client.http.headers["User-Agent"] == "Vesta"

    def test_compose(self, vbml_client: VBMLClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(vbml_client, json=chars)

        component = Component("template")
        props = {"prop": "value"}
        assert vbml_client.compose([component], props=props) == chars

        assert mock["url"] == "/compose"
        assert mock["json"] == {
            "components": [component.asdict()],
            "props": props,
        }

    def test_component_no_components(self, vbml_client: VBMLClient):
        with pytest.raises(ValueError, match=r"expected at least one component"):
            vbml_client.compose([])

    def test_format(self, vbml_client: VBMLClient, mock_response):
        chars = [[0] * COLS] * ROWS
        mock = mock_response(vbml_client, json=chars)

        assert vbml_client.format("message") == chars
        assert mock["url"] == "/format"



from __future__ import annotations

import http.client
import io
import urllib.error

import pytest

from vesta.http import Client
from vesta.http import HTTPError
from vesta.http import Response


@pytest.fixture
def mock_response(monkeypatch):
    """Fixture that returns a factory for mocking successful HTTP responses."""

    def _mock(client, status_code=200, body=b"{}"):
        last_request = {}

        def fake_open(request):
            last_request.update(
                {
                    "method": request.get_method(),
                    "url": request.get_full_url(),
                    "data": request.data,
                    "headers": dict(request.headers),
                }
            )

            class FakeSocket:
                def __init__(self, response_bytes):
                    self._file = io.BytesIO(response_bytes)

                def makefile(self, mode):
                    return self._file

            response_text = (
                f"HTTP/1.1 {status_code} OK\r\nContent-Length: {len(body)}\r\n\r\n"
            ).encode() + body

            sock = FakeSocket(response_text)
            resp = http.client.HTTPResponse(sock)
            resp.begin()

            return resp

        monkeypatch.setattr(client._opener, "open", fake_open)
        return last_request

    return _mock


class TestClient:
    def test_success(self, mock_response):
        client = Client(base_url="https://api.example.com")
        mock = mock_response(client, status_code=200, body=b'{"result":"ok"}')

        resp = client.request("GET", "/test")

        assert resp.status == 200
        assert resp.json() == {"result": "ok"}
        assert mock["url"] == "https://api.example.com/test"

    def test_http_error(self, monkeypatch):
        """Test that urllib.error.HTTPError is raised and propagated correctly."""
        client = Client(base_url="https://api.example.com")

        def raise_error(request):
            raise urllib.error.HTTPError(
                request.get_full_url(), 404, "Not Found", {}, None
            )

        monkeypatch.setattr(client._opener, "open", raise_error)

        with pytest.raises(HTTPError) as exc_info:
            client.request("GET", "/not-found")

        assert exc_info.value.code == 404
        assert exc_info.value.url == "https://api.example.com/not-found"


class TestResponse:
    def test_json_success(self):
        resp = Response(200, b'{"key": "value"}', {})
        assert resp.json() == {"key": "value"}

    def test_json_decode_error(self):
        resp = Response(200, b"not json", {})
        assert resp.json() is None
        assert resp.json(default={}) == {}
        assert resp.json(default=[]) == []
